/*
 * Copyright (c) 2024 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <config.h>

#ifndef HAVE_GETENTROPY

#include <sys/types.h>
#include <sys/time.h>
#ifdef HAVE_PRNGD
# include <sys/socket.h>
# include <sys/un.h>
# include <netinet/in.h>
#endif /* HAVE_PRNGD */
#ifdef HAVE_LINUX_RANDOM_H
# include <sys/syscall.h>
# include <linux/types.h>
# include <linux/random.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#if defined(TIME_WITH_SYS_TIME) || !defined(HAVE_SYS_TIME_H)
# include <time.h>
#endif

#include <extern.h>

#define minimum(a, b)	((a) < (b) ? (a) : (b))

#ifdef HAVE_PRNGD

#ifndef INADDR_LOOPBACK
# define INADDR_LOOPBACK ((unsigned int)0x7f000001)
#endif

static int
open_prngd(int len)
{
	int fd = -1, salen;
	char buf[2];
	union {
		struct sockaddr_in in;
		struct sockaddr_un un;
	} sa_un;
	struct sockaddr *sa = (struct sockaddr *)&sa_un;

	memset(&sa_un, 0, sizeof(sa_un));
#ifdef PRNGD_PORT
	sa_un.in.sin_family = AF_INET;
	sa_un.in.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	sa_un.in.sin_port = htons(PRNGD_PORT);
	salen = sizeof(sa_un.in);
#else /* PRNGD_PATH */
	sa_un.un.sun_family = AF_UNIX;
	strncpy(sa_un.un.sun_path, PRNGD_PATH, sizeof(sa_un.un.sun_path));
	salen = sizeof(sa_un.un) -
	    sizeof(sa_un.un.sun_path) + strlen(sa_un.un.sun_path);
#endif
	if ((fd = socket(sa->sa_family, SOCK_STREAM, 0)) == -1)
		goto bad;
	if (connect(fd, sa, salen) == -1)
		goto bad;
	signal(SIGPIPE, SIG_IGN);

	buf[0] = 0x02;
	buf[1] = len;
	if (write(fd, buf, 2) != 2)
		goto bad;

	return(fd);
bad:
	if (fd != -1)
		close(fd);
	return(-1);
}
#endif /* HAVE_PRNGD */

#if defined(_PATH_RANDOM) || defined(HAVE_PRNGD)
static int
getentropy_urandom(void *buf, size_t len)
{
	size_t offset = 0;
        int fd = -1;
        int ret = -1;

#ifdef _PATH_RANDOM
        fd = open(_PATH_RANDOM, O_RDONLY);
#endif
#ifdef HAVE_PRNGD
	if (fd == -1)
		fd = open_prngd(sizeof(rnd));
#endif
        if (fd != -1) {
		for (;;) {
			ssize_t nread = read(fd, buf + offset, len - offset);
			if (nread == -1) {
				if (errno == EINTR || errno == EAGAIN)
					continue; /* XXX - poll on EAGAIN */
				break;
			}
			offset += nread;
			if (offset == len) {
				 ret = 0;
				 break;
			}
		}
		close(fd);
        }

	return ret;
}
#else
static int
getentropy_urandom(void *buf, size_t len)
{
	return -1;
}
#endif /* _PATH_RANDOM || HAVE_PRNGD */


#if defined(SYS_getrandom) && defined(GRND_NONBLOCK)
static int
getentropy_getrandom(void *buf, size_t len)
{
	long ret;

	for (;;) {
		ret = syscall(SYS_getrandom, buf, len, GRND_NONBLOCK);
		if (ret != -1 || errno != EINTR)
			break;
	}
	if (ret < 0 || (size_t)ret != len) {
		/* Should not get short read from getrandom if non-blocking. */
		return -1;
	}
	return 0;
}
#else
static int
getentropy_getrandom(void *buf, size_t len)
{
	return -1;
}
#endif /* SYS_getrandom && GRND_NONBLOCK */

static int
getentropy_fallback(void *buf, size_t len)
{
	pid_t pid = getpid();
	size_t offset = 0;

	for (;;) {
		struct timeval tv;
		size_t copylen;
		int seed[4];

		/* Don't have random source, do our best... */
		gettimeofday(&tv, NULL);

		/* Use time of day and process id multiplied by small primes */
		seed[0] = (pid++ % 1000) * 983;
		seed[1] = tv.tv_usec * 13;
		seed[2] = (tv.tv_sec % 10000) * 523;
		seed[3] = (tv.tv_sec >> 10) * 389;

		copylen = minimum(len - offset, sizeof(seed));
		memcpy(buf + offset, seed, copylen);
		offset += copylen;
		if (offset == len)
			 return 0;
	}
}

int
repl_getentropy(void *buf, size_t len)
{
	int ret;

	ret = getentropy_getrandom(buf, len);
	if (ret != -1)
		return ret;

	ret = getentropy_urandom(buf, len);
	if (ret != -1)
		return ret;

	ret = getentropy_fallback(buf, len);

	return ret;
}
#endif /* HAVE_GETENTROPY */
