/*	$OpenBSD: mktemp.c,v 1.26 2024/03/01 21:50:40 millert Exp $	*/

/*
 * Copyright (c) 1996, 1997, 2001-2004, 2008, 2010, 2013, 2024
 *	Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <config.h>

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef HAVE_PATHS_H
#include <paths.h>
#endif /* HAVE_PATHS_H */
#ifdef HAVE_GETOPT_LONG
# include <getopt.h>
#else
# include "compat/getopt.h"
#endif /* HAVE_GETOPT_LONG */

#include <extern.h>

#ifndef _PATH_TMP
#define _PATH_TMP "/tmp"
#endif

#ifdef HAVE_PROGNAME
extern char *__progname;
#else
char *__progname;
#endif

__dead void usage(void);
__dead void fatal(const char *, ...) __printflike(1, 2);
__dead void fatalx(const char *, ...) __printflike(1, 2);

static struct option const longopts[] =
{
  {"directory",	no_argument,		NULL,	'd'},
  {"help",	no_argument,		NULL,	'h'},
  {"quiet",	no_argument,		NULL,	'q'},
  {"tmpdir",	optional_argument,	NULL,	'T'},
  {"dry-run",	no_argument,		NULL,	'u'},
  {"version",	no_argument,		NULL,	'V'},
  {NULL,	0,			NULL,	0}
};

static int quiet;

int
main(int argc, char *argv[])
{
	int ch, fd, uflag = 0, tflag = 0, Tflag = 0, makedir = 0;
	char *base, *cp, *template, *tempfile, *prefix = _PATH_TMP;
	size_t len, suffixlen = 0;

#ifndef HAVE_PROGNAME
	__progname = argv[0];
#endif

	while ((ch = getopt_long(argc, argv, "dp:qtuV", longopts, NULL)) != -1)
		switch (ch) {
		case 'd':
			makedir = 1;
			break;
		case 'p':
			prefix = optarg;
			tflag = 1;
			break;
		case 'q':
			quiet = 1;
			break;
		case 'T':
			if (optarg) {
				Tflag = 1;
				prefix = optarg;
			}
			/* FALLTHROUGH */
		case 't':
			tflag = 1;
			break;
		case 'u':
			uflag = 1;
			break;
		case 'V':
			printf("%s version %s\n", __progname, PACKAGE_VERSION);
			exit(0);
		default:
			usage();
	}

	/* If no template specified use a default one (implies -t mode) */
	switch (argc - optind) {
	case 1:
		template = argv[optind];
		break;
	case 0:
		template = "tmp.XXXXXXXXXX";
		tflag = 1;
		break;
	default:
		usage();
	}

	base = strrchr(template, '/');
	if (base != NULL)
		base++;
	else
		base = template;
	len = strlen(base);
	if (len > 0 && base[len - 1] != 'X') {
		/* Check for suffix, e.g. /tmp/XXXXXX.foo in last component. */
		for (suffixlen = 0; suffixlen < len; suffixlen++) {
			if (base[len - suffixlen - 1] == 'X')
				break;
		}
	}

	if (len - suffixlen < 6 ||
	    strncmp(&base[len - suffixlen - 6], "XXXXXX", 6)) {
		fatalx("insufficient number of Xs in template `%s'",
		    template);
	}
	if (tflag) {
		size_t size;

		if (base != template) {
			fatalx("template must not contain directory "
			    "separators in -t mode");
		}

		if (!Tflag) {
			cp = getenv("TMPDIR");
			if (cp != NULL && *cp != '\0')
				prefix = cp;
		}
		len = strlen(prefix);
		while (len != 0 && prefix[len - 1] == '/')
			len--;

		size = len + 1 + strlen(template) + 1;
		if ((tempfile = malloc(size)) != NULL) {
			if (snprintf(tempfile, size, "%.*s/%s", (int)len,
			    prefix, template) >= size) {
				free(tempfile);
				tempfile = NULL;
			}
		}
	} else {
		tempfile = strdup(template);
	}

	if (tempfile == NULL)
		fatalx("cannot allocate memory");

	if (makedir) {
		if (priv_mkdtemps(tempfile, suffixlen) == NULL)
			fatal("cannot make temp dir %s", tempfile);
		if (uflag)
			(void)rmdir(tempfile);
	} else {
		if ((fd = priv_mkstemps(tempfile, suffixlen)) == -1)
			fatal("cannot make temp file %s", tempfile);
		(void)close(fd);
		if (uflag)
			(void)unlink(tempfile);
	}

	(void)puts(tempfile);
	free(tempfile);

	return EXIT_SUCCESS;
}

__dead void
fatal(const char *fmt, ...)
{
	if (!quiet) {
		int saved_errno = errno;
		va_list ap;

		(void)fputs(__progname, stderr);
		(void)fputc(':', stderr);
		(void)fputc(' ', stderr);
		va_start(ap, fmt);
		(void)vfprintf(stderr, fmt, ap);
		va_end(ap);
		(void)fputc(':', stderr);
		(void)fputc(' ', stderr);
		(void)fputs(strerror(saved_errno), stderr);
		(void)fputc('\n', stderr);
	}
	exit(EXIT_FAILURE);
}

__dead void
fatalx(const char *fmt, ...)
{
	if (!quiet) {
		va_list ap;

		(void)fputs(__progname, stderr);
		(void)fputc(':', stderr);
		(void)fputc(' ', stderr);
		va_start(ap, fmt);
		(void)vfprintf(stderr, fmt, ap);
		va_end(ap);
		(void)fputc('\n', stderr);
	}
	exit(EXIT_FAILURE);
}

__dead void
usage(void)
{
	(void)fprintf(stderr,
	    "usage: %s [-V] | [-dqtu] [-p directory] [template]\n",
	    __progname);
	exit(EXIT_FAILURE);
}
